@@ -16,8 +16,16 @@ import pandas as pd
 import requests
 import base64
 from typing import List, Tuple, Optional
+from collections import OrderedDict
+import sys
+sys.path.append("..")
+from custom_exceptions.salesforce_exceptions import *
+from utilities.df_utils import *
+
 
 class Sf:
+    _REGEX_PARSE_SOQL_FIELDS = "(?<=select)(.*?)(?<!, )(?=from)"
+    _REGEX_PARSE_SOQL_OBJECT = "(?<!, from )(?<=from )\w*"
     def __init__(self) -> None:
         self.filename: str = None
         self.os_platform: str = platform
@@ -139,7 +147,7 @@ class Sf:
         bulk = SalesforceBulk(sessionId = session_id, host = instance)
         self.bulk = bulk
         # SFDC Metadata API
-        sfMeta = SfdcSession(session_id=session_id, instance=instance)
+        sfMeta = SfdcSession(session_id=session_id, instance=".".join(instance.split(".")[:2]))
         self.sfMeta = sfMeta
         # Alternative way to authenticate using SFDC Metadata API
         # sfMeta = SfdcSession(username=sfUsername,password=sfPassword,token='',is_sandbox=isSandbox)
@@ -191,7 +199,7 @@ class Sf:
             A Pandas Dataframe of the last created 100 account records, with all queriable fields included in the query and a relationship field.
             
         """
-        objectName = re.search('(?<=from )\w*', query.lower()).group(0)
+        objectName = re.search(self._REGEX_PARSE_SOQL_OBJECT, query.lower()).group(0)
         successful = False
         
         extracted_object = pd.DataFrame()
@@ -211,14 +219,16 @@ class Sf:
                         schemaDict[field]['name'] not in excludedFields):
                         results.append(field)
                 final_query = query.replace("*", ", ".join(results))                
-#                 # if the object has 0 records in Salesforce, return empty dataframe
-#                 if objectName not in self.record_count_caseinsensitive.keys():
-#                     return pd.DataFrame(columns=results)
                 query_response = self.sf.query_all(final_query)
-                result = pd.DataFrame(query_response)['records']
-                for _ in result:
-                    del _['attributes']
-                successful = True
+                # if the object has 0 records in Salesforce, return empty dataframe
+                if query_response['totalSize'] == 0:
+                    fields_preparsed = re.search(self._REGEX_PARSE_SOQL_FIELDS, query,  re.IGNORECASE).group(0).split(",")
+                    return pd.DataFrame(columns=fields_preparsed)
+                else:
+                    result = pd.DataFrame(query_response)['records']
+                    for _ in result:
+                        del _['attributes']
+                    successful = True
             except KeyError:
                 if self.debug:
                     print(objectName + ' skipped. (Potentially due to no records found.)')
@@ -237,7 +247,7 @@ class Sf:
             
             result = result.apply(lambda x: pd.Series(x)).copy()
 
-            relationship_fields_preparsed = re.search('(?<=select )(.*?)(?= from)', query,  re.IGNORECASE).group(0).split(",")
+            relationship_fields_preparsed = re.search(self._REGEX_PARSE_SOQL_FIELDS, query,  re.IGNORECASE).group(0).split(",")
             # relationship_fields_prepared Example:
             # ['*',
             #  ' Parent_Account_vod__r.Owner.Profile.Name',
@@ -280,6 +290,149 @@ class Sf:
                 
         return result
 
+    def create(self, object_api: str=None, record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
+        """
+        Function creates records within the Salesforce instance using bulk api.
+
+        Arguments:
+        object_api: str: API name of the Salesforce Object for which the records are to be created.
+
+        record_dataframe: pd.DataFrame: A pandas dataframe object containing all the required fields for a record.
+
+        Returns:
+        result: pd.DataFrame: A pandas dataframe with the following columns:
+            success: boolean: indicates whether the create request was successful
+            created: boolean: this value should always be False in a create request
+            id: object[str]: this value typically is None in a create request
+            statusCode: object[str] - Optional: the error code of the create operation if failed
+            message: object[str] - Optional: the error message of the create operation if failed
+            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.
+
+        Raises:
+            RequiredValuesNotProvidedDuringUpdate
+        """
+        if (object_api is None) or (len(record_dataframe) == 0):
+            raise RequiredValuesNotProvidedDuringCreate()
+        else:
+            result = getattr(self.sf.bulk, object_api).insert(record_dataframe.to_dict('records'))
+            result = pd.DataFrame(result)
+            result = unpack_column(result, "errors")
+
+        return pd.DataFrame(result.rename(columns={'id': 'Id'}))
+
+    def delete(self, object_api: str=None, record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
+        """
+        Function deletes records within the Salesforce instance using bulk api.
+
+        Arguments:
+        object_api: str: API name of the Salesforce Object for which the records are to be deleted from.
+
+        record_dataframe: pd.DataFrame: A pandas dataframe object containing at least 1 column with an Id column (case sensitive)
+
+        Returns:
+        result: pd.DataFrame: A pandas dataframe with the following columns:
+            success: boolean: indicates whether the delete request was successful
+            created: boolean: this value should always be False in a delete request
+            id: object[str]: this value typically is None in a delete request
+            statusCode: object[str] - Optional: the error code of the delete operation if failed
+            message: object[str] - Optional: the error message of the delete operation if failed
+            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.
+
+        Raises:
+            RequiredValuesNotProvidedDuringDelete
+        """
+        if (object_api is None) or (len(record_dataframe) == 0) or record_dataframe.columns.__contains__('Id') == False:
+            raise RequiredValuesNotProvidedDuringDelete()
+        else:
+            result = getattr(self.sf.bulk, object_api).delete(record_dataframe['Id'].to_frame().to_dict('records'))
+            result = pd.DataFrame(result)
+            result = unpack_column(result, "errors")
+
+        return pd.DataFrame(result.rename(columns={'id': 'Id'}))
+    def update(self, object_api: str=None, record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
+        """
+        Function updates records within the Salesforce instance using bulk api.
+
+        Arguments:
+        object_api: str: API name of the Salesforce Object for which the records are to be updated.
+
+        record_dataframe: pd.DataFrame: A pandas dataframe object containing at least 1 column with an Id column (case sensitive)
+            and additional columns matching the Salesforce field api names.
+
+        Returns:
+        result: pd.DataFrame: A pandas dataframe with the following columns:
+            success: boolean: indicates whether the update request was successful
+            created: boolean: this value should always be False in a update request
+            id: object[str]: this value typically is None in a update request
+            statusCode: object[str] - Optional: the error code of the update operation if failed
+            message: object[str] - Optional: the error message of the update operation if failed
+            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.
+
+        Raises:
+            RequiredValuesNotProvidedDuringUpdate
+        """
+
+        field_metadata = self.field_describe([object_api])
+
+        if ((object_api is None) or 
+        (len(record_dataframe) == 0) or 
+        record_dataframe.columns.__contains__('Id') == False or
+        # checks whether any of the columns in the dataframe passed in are not valid field API names
+        ~record_dataframe.columns.isin(field_metadata[object_api]).all()
+        # checks whether an reference (lookup/master-detail) field is included in the source dataframe
+        #  or ~record_dataframe.columns.isin(field_metadata[field_metadata['Type'] == 'reference'][object_api]).any()
+        ):
+            raise RequiredValuesNotProvidedDuringUpdate()
+        else:
+            result = getattr(self.sf.bulk, object_api).update(record_dataframe.to_dict('records'))
+            result = pd.DataFrame(result)
+            result = unpack_column(result, "errors")
+
+        return pd.DataFrame(result.rename(columns={'id': 'Id'}))
+
+    def upsert(self, object_api: str=None, external_id_field_api: str="", record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
+        """
+        Function upserts records within the Salesforce instance using bulk api.
+
+        Arguments:
+        object_api: str: API name of the Salesforce Object for which the records are to be upserted into.
+
+        external_id_field_api: str: API name of the external ID field used for the upsert operation.
+
+        record_dataframe: pd.DataFrame: A pandas dataframe object containing the External column (case sensitive)
+            and additional columns matching the Salesforce field api names.
+
+        Returns:
+        result: pd.DataFrame: A pandas dataframe with the following columns:
+            success: boolean: indicates whether the upsert request was successful
+            created: boolean: this value should always be False in a upsert request
+            id: object[str]: this value typically is None in a upsert request
+            statusCode: object[str] - Optional: the error code of the upsert operation if failed
+            message: object[str] - Optional: the error message of the upsert operation if failed
+            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.
+
+        Raises:
+            RequiredValuesNotProvidedDuringUpsert
+            SalesforceMalformedRequest
+        """
+
+        field_metadata = self.field_describe([object_api], ['name', 'type', 'length','externalId'])
+
+        if ((object_api is None) or 
+        (len(record_dataframe) == 0) or 
+        # record_dataframe.columns.__contains__('Id') == False or
+        # checks whether any of the columns in the dataframe passed in are not valid field API names
+        ~record_dataframe.columns.isin(field_metadata[object_api]).all()
+         or external_id_field_api==""
+        ):
+            raise RequiredValuesNotProvidedDuringUpsert()
+        else:
+            result = getattr(self.sf.bulk, object_api).upsert(record_dataframe.to_dict('records'), external_id_field_api)
+            result = pd.DataFrame(result)
+            result = unpack_column(result, "errors")
+
+        return pd.DataFrame(result)
+
     def load_credentials(self,
                                 os_platform: Optional[str] = None, 
                                 filename: Optional[str] = None, 
@@ -350,7 +503,7 @@ class Sf:
             import pandas as pd
         
         """
-        objectName = re.search('(?<=from )\w*', og_query.lower()).group(0)
+        objectName = re.search(self._REGEX_PARSE_SOQL_OBJECT, og_query.lower()).group(0)
         successful = False
         
         extracted_object = pd.DataFrame()
@@ -420,7 +573,6 @@ class Sf:
             print("Extracted " + objectName + " successfully!")
         return sfdf
 
-
     def join(self, 
                 dataframe, 
                 rt_dataframe, 
@@ -440,21 +592,85 @@ class Sf:
 #         dataframe.drop([col for col in dataframe.columns if 'drop' in col], axis=1, inplace=True)
         return dataframe
 
-    def object_describe(self, sobject_api_name: str):
+    def object_describe(self, sobject_api_name: str, export_excel: bool=False):
         """
-        TODO:
-        
-        Description of what it does
-        
-        Description of arguments and data types
-        
-        Description of return values and data types
-        
-        Description of Errors raised
-        
-        Extra Notes and Examples of Usage
+
         """
-        return getattr(self.sf, sobject_api_name).describe()
+        sfSchema = getattr(self.sf, sobject_api_name).describe()
+        boolMetadata = {}
+        nonetypeMetadata = {}
+        strMetadata = {}
+        orderedDictMetadata = pd.Series(dtype='object')
+        childRelationshipsPD = pd.DataFrame()
+        recordTypeInfosPD = pd.DataFrame()
+        fieldsPD = pd.DataFrame()
+        supportedScopesPD = pd.DataFrame()
+        namedLayoutInfosPD = pd.DataFrame()
+        actionOverridesPD = pd.DataFrame()
+
+        for metadata in sfSchema:
+            ## Processes all metadata that would return a boolean datatype
+            if type(sfSchema[metadata]) is bool:
+                boolMetadata[metadata] = sfSchema[metadata]
+
+            ## Processes all metadata that would return a NoneType datatype
+            elif type(sfSchema[metadata]) == type(None):
+                nonetypeMetadata[metadata] = sfSchema[metadata]
+
+            ## Processes all metadata that would return a string datatype
+            elif type(sfSchema[metadata]) is str:
+                strMetadata[metadata] = sfSchema[metadata]
+            ## Processes all metadata that would return an OrderedDict data type
+            elif type(sfSchema[metadata]) is OrderedDict and len(sfSchema[metadata]) != 0:
+                orderedDictMetadata = pd.concat([orderedDictMetadata,pd.Series(sfSchema['urls'])])
+            elif type(sfSchema[metadata]) is OrderedDict and len(sfSchema[metadata]) == 0:
+                nonetypeMetadata[metadata] = "None"
+
+            ## Processes all metadata that would return a list data type
+            elif type(sfSchema[metadata]) is list and len(sfSchema[metadata]) != 0:
+                if metadata == 'childRelationships':
+                    childRelationshipsPD = pd.DataFrame(sfSchema['childRelationships'], index = pd.DataFrame(sfSchema['childRelationships'])['field'])
+                    childRelationshipsPD = childRelationshipsPD.drop('field', axis=1).T
+                if metadata == 'recordTypeInfos':
+                    recordTypeInfosPD = pd.DataFrame(sfSchema['recordTypeInfos'], index = pd.DataFrame(sfSchema['recordTypeInfos'])['name'])
+                    recordTypeInfosPD = recordTypeInfosPD.drop('name', axis = 1).T  
+                if metadata == 'fields':
+                    fieldsPD = pd.DataFrame(sfSchema['fields'], index = pd.DataFrame(sfSchema['fields'])['name'])
+                    fieldsPD = fieldsPD.drop('name', axis = 1).T
+                if metadata == 'supportedScopes':
+                    supportedScopesPD = pd.DataFrame(sfSchema['supportedScopes'], index = pd.DataFrame(sfSchema['supportedScopes'])['name'])
+                    supportedScopesPD = supportedScopesPD.drop('name', axis = 1).T
+                if metadata == 'actionOverrides':
+                    actionOverridesPD = pd.DataFrame(sfSchema['actionOverrides'], index = pd.DataFrame(sfSchema['actionOverrides'])['name'])
+                    actionOverridesPD = actionOverridesPD.drop('name', axis = 1).T
+                if metadata == 'namedLayoutInfos':
+                    namedLayoutInfosPD = pd.DataFrame(sfSchema['namedLayoutInfos'], index = pd.DataFrame(sfSchema['namedLayoutInfos'])['name'])
+                    namedLayoutInfosPD = namedLayoutInfosPD.drop('name', axis = 1).T
+            ## Parses all empty list metadata
+            elif type(sfSchema[metadata]) is list and len(sfSchema[metadata]) == 0:
+                nonetypeMetadata[metadata] = "None"
+            else:
+                nonetypeMetadata[metadata] = "Unrecognized metadata type: " + metadata
+
+        output = {'metadata': pd.concat([pd.Series(nonetypeMetadata),
+                    pd.Series(boolMetadata),
+                    pd.Series(strMetadata),
+                    orderedDictMetadata]),
+                    'childRelationships': childRelationshipsPD if len(childRelationshipsPD)!=0 else None,
+                    'recordTypeInfos': recordTypeInfosPD if len(recordTypeInfosPD)!=0 else None,
+                    'fields': fieldsPD if len(fieldsPD)!=0 else None,
+                    'supportedScopes': supportedScopesPD if len(supportedScopesPD)!=0 else None, 
+                    'namedLayoutInfos': namedLayoutInfosPD if len(namedLayoutInfosPD)!=0 else None,
+                    'actionOverrides': actionOverridesPD if len(actionOverridesPD)!=0 else None}
+        if export_excel:
+            with pd.ExcelWriter(sobject_api_name + ' describe output.xlsx') as writer:  
+                for key in output.keys():
+                    try:
+                        pd.DataFrame(output[key]).to_excel(writer, sheet_name=key)
+                    except:
+                        continue
+
+        return output
 
     def field_describe(self, objects: List = ['Account','Address_vod__c','Child_Account_vod__c'], 
     attributes: List = ['name','type','length']) -> pd.DataFrame:
